# vim: set fileencoding=<utf-8> :
# Copyright 2018-2020 John Lees and Nick Croucher

'''PopPUNK (POPulation Partitioning Using Nucleotide Kmers)'''

__version__ = '2.3.0'

# Minimum sketchlib version
SKETCHLIB_MAJOR = 1
SKETCHLIB_MINOR = 6
SKETCHLIB_PATCH = 0
